# pylint: disable=no-self-use
import json
import time
from datetime import datetime
from urllib.error import URLError

import pandas as pd
import requests
from dateutil import parser

from sdc_dp_helpers.api_utilities.date_managers import date_handler
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler


class CustomOneSignalReader:
    def __init__(self, creds_file, config_file, **kwargs):
        self._creds = load_file(creds_file, "yml")
        self._config = load_file(config_file, "yml")

        self._header = {
            "Content-Type": "application/json; charset=utf-8",
            "Authorization": f"Basic {self._creds.get('api_key')}",
            "User-Agent": "Mozilla/5.0",
        }

        self._request_session = requests.Session()
        self.offset, self.data_set = 0, []

        self._csv_wait_time = kwargs.get("csv_wait_time", 90)
        self._csv_get_attempts = kwargs.get("csv_get_attempts", 10)

    def _generate_url(self, endpoint: str):
        """
        Generate url string for authentication from given endpoints.
        :param endpoint: view_notification or csv_export
        :return: Url string.
        """
        _app_id = self._creds.get("app_id")
        if endpoint == "view_notifications":
            return f"https://onesignal.com/api/v1/notifications?app_id={_app_id}"
        if endpoint == "csv_export":
            return f"https://onesignal.com/api/v1/players/csv_export?app_id={_app_id}"

        raise ValueError(
            f"Given endpoint: {endpoint} is not supported. "
            f"Try csv_export or view_notifications."
        )

    @request_handler(wait=0.5, backoff_factor=0.5)
    def _view_notification_query_handler(self, limit: int, offset: int):
        """
        Handles the view notification request attempt.
        """
        url = self._generate_url(endpoint="view_notifications")
        return requests.get(
            url=url,
            headers=self._header,
            params={"limit": limit, "total_count": "true", "offset": offset},
        )

    @request_handler(wait=1, backoff_factor=0.5)
    def _csv_export_query_handler(self):
        """
        Handles the csv export request attempt.
        """
        url = self._generate_url(endpoint="csv_export")
        return requests.post(url=url, headers=self._header)

    def _get_csv_export_handler(self, response):
        """
        Attempts to gather the data from the csv url, and waits while
        it is being generated.
        :param response: Onesignal response object.
        """
        csv_url = response.get("csv_file_url", None)
        if csv_url is not None:
            attempts = 1
            while True:
                try:
                    print(f"Attempting to gather data from url: {csv_url}.")
                    data_frame: pd.DataFrame = pd.read_csv(csv_url)
                    self.data_set = data_frame.to_dict(orient="records")

                    # create a usable field to write to s3
                    # also handle timestamps that have milliseconds, it's not
                    # always there, so we need to handle both
                    for row in self.data_set:
                        try:
                            row["created_at_date"] = str(
                                datetime.strptime(
                                    row["created_at"], "%Y-%m-%d %H:%M:%S"
                                ).date()
                            ).replace("-", "")
                        except ValueError:
                            row["created_at_date"] = str(
                                datetime.strptime(
                                    row["created_at"], "%Y-%m-%d %H:%M:%S.%f"
                                ).date()
                            ).replace("-", "")

                    print("Data gathered...")
                    return self.data_set

                except URLError:
                    print(
                        f"CSV file not generated, waiting for {self._csv_wait_time} seconds. "
                        f"Attempt {attempts}/{self._csv_get_attempts}."
                    )
                    time.sleep(self._csv_wait_time)
                    if attempts >= self._csv_get_attempts:
                        raise ConnectionError(
                            f"CSV file failed to generate after {attempts} attempts. "
                            "Contact Onesignal for help."
                        )
                    attempts += 1
        else:
            raise ConnectionError(response)

    def run_query(self):
        """
        Generate a compressed CSV export of all of your current user data.
        This method can be used to generate a compressed CSV export of all of your current user data.
        View the details of multiple notifications.
        :return: The response dataset.
        """
        _endpoint = self._config.get("endpoint")
        _limit = self._config.get("limit", 50)
        print(f"Gathering data for {_endpoint}.")

        if _endpoint == "view_notifications":
            # gather data per offset given the set of limits
            initial_response = self._view_notification_query_handler(
                limit=_limit, offset=0
            ).json()
            _total_records = int(initial_response.get("total_count"))

            for offset in range(0, _total_records, _limit):
                print(f"At offset {offset} of {_total_records}.")
                response = self._view_notification_query_handler(
                    limit=_limit, offset=offset
                ).json()
                for notification in response.get("notifications"):
                    self.data_set.append(notification)

            return self.data_set

        if _endpoint == "csv_export":
            # if connection is good, try to get the csv file from the Onesignal
            # Athena wrapper, it could take time to generate, so wait for it.
            response = self._csv_export_query_handler().json()
            self._get_csv_export_handler(response=response)
            return self.data_set

        raise ValueError(
            f"Given endpoint: {_endpoint} is not supported. "
            f"Try csv_export or view_notifications."
        )



# !/usr/bin/env python
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates refresh token for AdWords using the Installed Application flow."""

import argparse
import sys

from google_auth_oauthlib.flow import InstalledAppFlow
from oauthlib.oauth2.rfc6749.errors import InvalidGrantError

# Your OAuth2 Client ID and Secret. If you do not have an ID and Secret yet,
# please go to https://console.developers.google.com and create a set.
DEFAULT_CLIENT_ID = None
DEFAULT_CLIENT_SECRET = None

# The AdWords API OAuth2 scope.
SCOPE = "https://www.googleapis.com/auth/adwords"
# The redirect URI set for the given Client ID. The redirect URI for Client ID
# generated for an installed application will always have this value.
_REDIRECT_URI = "urn:ietf:wg:oauth:2.0:oob"

parser = argparse.ArgumentParser(
    description="Generates a refresh token with " "the provided credentials."
)
parser.add_argument(
    "--client_id",
    default=DEFAULT_CLIENT_ID,
    help="Client Id retrieved from the Developer's Console.",
)
parser.add_argument(
    "--client_secret",
    default=DEFAULT_CLIENT_SECRET,
    help="Client Secret retrieved from the Developer's " "Console.",
)
parser.add_argument(
    "--additional_scopes",
    default=None,
    help="Additional scopes to apply when generating the "
    "refresh token. Each scope should be separated by a comma.",
)


class ClientConfigBuilder(object):
    """Helper class used to build a client config dict used in the OAuth 2.0 flow."""

    _DEFAULT_AUTH_URI = "https://accounts.google.com/o/oauth2/auth"
    _DEFAULT_TOKEN_URI = "https://accounts.google.com/o/oauth2/token"
    CLIENT_TYPE_WEB = "web"
    CLIENT_TYPE_INSTALLED_APP = "installed"

    def __init__(
        self,
        client_type=None,
        client_id=None,
        client_secret=None,
        auth_uri=_DEFAULT_AUTH_URI,
        token_uri=_DEFAULT_TOKEN_URI,
    ):
        self.client_type = client_type
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_uri = auth_uri
        self.token_uri = token_uri

    def Build(self):
        """Builds a client config dictionary used in the OAuth 2.0 flow."""
        if all(
            (
                self.client_type,
                self.client_id,
                self.client_secret,
                self.auth_uri,
                self.token_uri,
            )
        ):
            client_config = {
                self.client_type: {
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "auth_uri": self.auth_uri,
                    "token_uri": self.token_uri,
                }
            }
        else:
            raise ValueError("Required field is missing.")

        return client_config


def main(client_id, client_secret, scopes):
    """Retrieve and display the access and refresh token."""
    client_config = ClientConfigBuilder(
        client_type=ClientConfigBuilder.CLIENT_TYPE_WEB,
        client_id=client_id,
        client_secret=client_secret,
    )

    flow = InstalledAppFlow.from_client_config(client_config.Build(), scopes=scopes)
    # Note that from_client_config will not produce a flow with the
    # redirect_uris (if any) set in the client_config. This must be set
    # separately.
    flow.redirect_uri = _REDIRECT_URI

    auth_url, _ = flow.authorization_url(prompt="consent")

    print(
        "Log into the Google Account you use to access your AdWords account "
        "and go to the following URL: \n%s\n" % auth_url
    )
    print("After approving the token enter the verification code (if specified).")
    code = input("Code: ").strip()

    try:
        flow.fetch_token(code=code)
    except InvalidGrantError as ex:
        print("Authentication has failed: %s" % ex)
        sys.exit(1)

    print("Access token: %s" % flow.credentials.token)
    print("Refresh token: %s" % flow.credentials.refresh_token)



# pylint: disable=too-few-public-methods

"""
    CUSTOM WRITER CLASS FOR XERO DATA
"""
import os
import json
import datetime
import boto3

class CustomS3JsonWriter:
    """Class to write files to s3"""
    data = None
    def __init__(self, bucket: str, profile_name: str=None) -> None:
        self.bucket = bucket
        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource("s3")

    def write_to_s3(self, json_data: json, config: dict) -> None:
        """
        Construct partitioning and file name conventions in s3
        according to business specifications, and write to S3.
        """
        data_variant = config.get("data_variant").lower()
        collection_name = config.get("collection_name").lower()
        organisation_name = config.get("tenant_name").lower()
        if not json_data:
            return
        if not ( data_variant and collection_name and organisation_name ):
            raise ValueError(
                'One or more required attributes ("data_variant", "collection_name", "tenant_name") missing from request config'
            )
        
        _date = config.get( "date", config.get("date", datetime.datetime.now().strftime("%Y-%m-%d") ) ).replace("-", "")

        key_path = f"{collection_name}/{data_variant}/{organisation_name}/{_date}.json"

        # in the case of reports we want to add contact_id as a partition_part, let's check for extra partition_parts
        partition_part = config.get("partition_part")
        if partition_part:
            key_path = key_path.replace(f"{organisation_name}/", f"{organisation_name}/{partition_part}/")

        # replace any '-' in the path, just in case; we don't want s3 to panic
        key_path = "_".join( key_path.split("-") )
        key_path = "_".join( key_path.split(" ") )

        print( f"Write path: S3://{self.bucket}/{key_path}" )
        self.data = json_data
        self.s3_resource.Object(self.bucket, key_path).put( Body = json.dumps( json_data ) )



# pylint: disable=too-few-public-methods

"""
    CUSTOM WRITER CLASSES
"""
import json
import os
from datetime import datetime

import boto3


class CustomFalconWriter:
    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.folder_path = folder_path

    def write_to_s3(self, payload: dict) -> None:
        """
        This pulls the yielded dataset from the GA reader in a manner
        that consumes the dataset of the given view_id and date,
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list(),
                            "date": string,
                            "view_id": string
                        }
        """
        _data: dict = payload["data"]
        _date: str = payload["date"].replace("-", "")
        _networks: str = payload["networks"]
        # _endpoint_name: str = payload["endpoint_name"]
        write_path: str = f"{self.folder_path}/{_networks}/{_date}.json"
        print(
            f"Writing data to s3://{self.bucket}/{write_path} partitioned by endpoint_name, networks and date."
        )
        self.s3_resource.Object(self.bucket, write_path).put(Body=json.dumps(_data))


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, file_path, profile_name=None):
        self.bucket = bucket
        self.file_path = file_path

        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource("s3")

    def write_to_s3(self, json_data):
        """
        Writes json data to s3 and names it by
        date of writing.
        """
        now = datetime.now().strftime("%Y_%m_%d_%H%M%S")
        key_path = f"{os.path.join(self.file_path, now)}.json"

        self.s3_resource.Object(self.bucket, key_path).put(Body=json.dumps(json_data))



"""
    CUSTOM FALCON READER CLASSES
"""
import os
from typing import Generator, Callable, Set
from datetime import datetime, timedelta

import requests

from sdc_dp_helpers.api_utilities.date_managers import date_handler, date_range
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class CustomFalconReader:
    """
    Custom Falcon Reader
    """

    def __init__(self, creds_file: str, config_file=None):
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")

        self.request_session: requests.Session = requests.Session()
        self.base_url: str = "https://api.falcon.io/"
        self.datasets: list = []

    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def _get_channel_ids(self) -> Set[str]:
        """
        Gather all available channel ids.
        """
        print("GET: channel ids.")
        url = f"https://api.falcon.io/channels?apikey={self._creds['api_key']}"
        response = self.request_session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = set()
            for item in response_data.get("items", []):
                channel_ids.add(item["id"])

            return channel_ids

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _content_by_channel_id(self, date: str, channel_id: str) -> list:
        """Gets the content metrics by channel id
        SEE: https://falconio.docs.apiary.io/#reference/channel-api/get-facebook-and-instagram-page-insights-metrics/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string pushed as '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """
        dataset: list = []
        endpoint_url: str = (
            f"measure/api/v1/content/metrics?apikey={self._creds['api_key']}"
        )
        offset: int = 0
        limit = self._config.get("limit", 200)
        start_date = datetime.strptime(date, "%Y-%m-%d").isoformat()
        end_date = (datetime.strptime(date, "%Y-%m-%d") + timedelta(days=1)).isoformat()
        while True:
            req = requests.Request(
                method=self._config.get("method", "POST"),
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                json={
                    "metrics": self._config.get("metrics", []),
                    "channels": [channel_id],
                    "since": start_date,
                    "until": end_date,
                    "postsSince": start_date,
                    "postsUntil": end_date,
                    "direction": self._config.get("direction", "ASC"),
                    "limit": limit,
                    "offset": offset,
                },
            )
            prepped = self.request_session.prepare_request(req)
            response = self.request_session.send(
                prepped,
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if status_code == 200:
                results: dict = response.json()
                offset += limit
                if len(results) == 0:
                    break  # we will exit the loop because no data is got back
                dataset.extend(results)

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_id}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _published_posts_by_channel_id(self, date: str, channel_id: str) -> list:
        """Gets the published posts by channel id
        SEE: https://falconio.docs.apiary.io/#reference/content-api/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string pushed as '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """

        dataset: list = []
        endpoint_url: str = f"publish/items?apikey={self._creds['api_key']}"
        limit = self._config.get("limit", 200)
        start_date = datetime.strptime(date, "%Y-%m-%d").isoformat()
        end_date = (datetime.strptime(date, "%Y-%m-%d") + timedelta(days=1)).isoformat()
        while endpoint_url:
            req = requests.Request(
                method=self._config.get("method", "GET"),
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                params={
                    "channels": [channel_id],
                    "since": start_date,
                    "until": end_date,
                    "networks": self._config["networks"],
                    "statuses": self._config.get("statuses", "published"),
                    "limit": limit,
                },
            )
            prepped = self.request_session.prepare_request(req)
            response = self.request_session.send(
                prepped,
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if response.status_code == 200:
                results: dict = response.json()
                items_data: list = results.get("items", [])
                dataset.extend(items_data)

                endpoint_url = results.get("next", {"href": None}).get("href")

                if len(items_data) == 0 or endpoint_url is None:
                    break

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_id}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    def _query_handler(self, date: str, channel_id: str) -> list:
        endpoint_name = self._config.get("endpoint_name", None)
        if not endpoint_name:
            raise Exception(
                "You need to specify endpoint in the configs i.e content_metrics or published_posts"
            )
        elif endpoint_name == "content_metrics":
            query = self._content_by_channel_id
        elif endpoint_name == "published_posts":
            query = self._published_posts_by_channel_id

        return query(date, channel_id)

    def run_query(self) -> Generator[dict, None, None]:
        """
        Get metrics by channel Id context returns a request session with the ability
        to page with offsets.
        Content (or Post level) contains all metrics about your specific piece of
        content (posts). Here you will find impressions, reach, likes,
        shares and other metrics that show how well your specific post has performed.
        https://falconio.docs.apiary.io/reference/content-api/get-copied-content.
        """
        channel_ids = self._get_channel_ids()
        # channel_ids = ["114215121990330", "58283981268"]
        print(f"Attempting to gather metrics from {len(channel_ids)} channel ids.")
        for date in date_range(
            start_date=date_handler(self._config.get("since", None), "%Y-%m-%d"),
            end_date=date_handler(self._config.get("until", None), "%Y-%m-%d"),
        ):
            payload: list = []
            for idx, channel_id in enumerate(channel_ids):
                data = self._query_handler(channel_id=channel_id, date=date)
                if data and len(data) > 0:
                    payload.extend(data)

            if len(payload) > 0:
                yield {
                    "networks": self._config["networks"],
                    "date": date.replace("-", ""),
                    "data": payload,
                }
            else:
                print(
                    f"No data for endpoint {self._config['endpoint_name']} for date : {date}"
                )
